// MIT License
//
// Copyright (c) 2019 Tobias Pfeiffer
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

//! # vkgen
//! Generates Rust source code from vk.xml
//!
//! ### Benefits
//! - easy to use
//! - no dependencies except libloading
//! - only ~1.500 lines of code
//! - can easily be modified to support other languages (just replace gen() with your own code)
//!
//! ### General information
//! - libloading is required to load vkGetInstanceProcAddr from the Vulkan shared library
//! - all other functions are loaded dynamically during runtime via vkGetInstanceProcAddr and vkGetDeviceProcAddr
//! - a very thin wrapper is generated for all dispatchable handles to store function pointers / parent handles,
//! but without any remarkable performance penalties
//!
//! ## Usage
//!
//! ### Step 1
//!
//! Download vk.xml from the official Vulkan-Headers repository (https://github.com/KhronosGroup/Vulkan-Headers/blob/master/registry/vk.xml)
//!
//! ### Step 2
//!
//! Generate Rust source code from vk.xml
//!
//! ```$ vkgen <input file> <flags>```
//!
//! ##### Parameters
//!
//! - `input file` is the Vulkan registry file to be parsed
//! - `flags` see Flags
//!
//! ##### Flags
//!
//! - `--verbose`, `-v` print all warnings
//!
//! ##### Output
//!
//! - `<input file>.rs` contains the generated source code
//! - `<input file>.toml` contains a list of all versions and extensions
//!
//! ##### Example
//!
//! ```$ vkgen ./vk.xml```
//!
//! Parses a file named `vk.xml` containing the Vulkan registry and outputs the generated
//! code to `vk.rs` and a list of all versions and extensions to `vk.toml`.
//!
//! ### Step 3
//!
//! Copy the Rust file into your project folder and add the versions and extensions located
//! in the toml file to your project's Cargo.toml:
//! ```toml
//! ...
//! [features]
//! default = ["VK_VERSION_1_0", "VK_VERSION_1_1"]
//! ...
//! VK_VERSION_1_0 = []
//! VK_VERSION_1_1 = []
//! ...
//! [dependencies]
//! libloading = "*"
//! ...
//! ```
//! ### Step 3a (optional)
//!
//! If you do not want to use libloading, remove the dependency and edit `vkInit` to load the
//! function pointers with your preferred method.
//!
//! ## Examples
//!
//! This simple example demonstrates how to load libvulkan on linux, output the instance
//! version (1.1.0) and create an instance. vk.rs is the file containing the generated Rust
//! source code.
//!
//! Cargo.toml:
//! ```toml
//! [package]
//! name = "vkgen_test"
//! version = "0.1.0"
//! authors = ["tobias"]
//! edition = "2018"
//!
//! [features]
//! default = ["VK_VERSION_1_0", "VK_VERSION_1_1"]
//! VK_VERSION_1_0 = []
//! VK_VERSION_1_1 = []
//!
//! [dependencies]
//! libloading = "0.5.0"
//! ```
//! main.rs:
//! ```rust
//! mod vk;
//!
//! use self::vk::*;
//! use std::ptr::null;
//!
//! fn main() {
//! 	let mut v: u32 = 0;
//! 	enumerateInstanceVersion(&mut v);
//! 	println!("vulkan instance version is {}.{}.{}", VK_VERSION_MAJOR(v), VK_VERSION_MINOR(v), VK_VERSION_PATCH(v));
//!
//! 	let instance_info = VkInstanceCreateInfo {
//! 		sType:                   VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
//! 		pNext:                   null(),
//! 		flags:                   0,
//! 		pApplicationInfo: &VkApplicationInfo {
//! 			sType:              VK_STRUCTURE_TYPE_APPLICATION_INFO,
//! 			pNext:              null(),
//! 			pApplicationName:   b"test app\0".as_ptr(),
//! 			applicationVersion: VK_MAKE_VERSION(0, 0, 1),
//! 			pEngineName:        b"test engine\0".as_ptr(),
//! 			engineVersion:      VK_MAKE_VERSION(0, 0, 1),
//! 			apiVersion:         VK_MAKE_VERSION(1, 1, 0),
//! 		},
//! 		enabledLayerCount:       0,
//! 		ppEnabledLayerNames:     null(),
//! 		enabledExtensionCount:   0,
//! 		ppEnabledExtensionNames: null()
//! 	};
//!
//! 	let mut instance = VK_NULL_HANDLE;
//! 	if createInstance(&instance_info, null(), &mut instance) != VK_SUCCESS {
//! 		panic!("something went wrong :-/");
//! 	};
//! 	let instance = unsafe { VkInstanceImpl::new(instance) };
//! }
//!
//! ```

#![allow(clippy::cognitive_complexity)]

use self::{State::*, ParseArgs::*, VkRegistryElement::*};
use std::{
	io::*,
	fs::{File, OpenOptions},
	collections::{HashSet, HashMap},
};
use std::convert::identity;

static VK_BEGINNING: &str = r#"//! MACHINE GENERATED FILE, DO NOT EDIT
#![allow(non_snake_case)]
#![allow(non_camel_case_types)]
#![allow(non_upper_case_globals)]
#![allow(dead_code)]
#![allow(clippy::unused_unit)]
#![allow(clippy::unreadable_literal)]
#![allow(clippy::too_many_arguments)]

use core::{mem::{transmute, MaybeUninit}, fmt::{Debug, Formatter, Error}};

pub type wl_display       = *mut ();
pub type wl_surface       = *mut ();
#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct xcb_connection_t;
#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct xcb_window_t;
#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct xcb_visualid_t;
pub type VisualID         = u32;
pub type Window           = u32;
pub type RROutput         = u32;
#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct Display;
pub type GgpStreamDescriptor = usize;
pub type GgpFrameToken       = usize;
pub type zx_handle_t         = usize;
#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct ANativeWindow;
#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct AHardwareBuffer;
#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct CAMetalLayer;
pub type HINSTANCE        = *mut ();
pub type HWND             = *mut ();
pub type HMONITOR         = *mut ();
pub type HANDLE           = *mut ();
pub type DWORD            = u32;
pub type LPCWSTR          = *const u16;
pub struct SECURITY_ATTRIBUTES {
	pub nLength:              DWORD,
	pub lpSecurityDescriptor: *mut (),
	pub bInheritHandle:       i32
}

pub const VK_NULL_HANDLE        : u64 = 0;
/// Vulkan 1.0 version number
pub const VK_API_VERSION_1_0    : u32 = VK_MAKE_VERSION(1, 0, 0);
/// Vulkan 1.1 version number
pub const VK_API_VERSION_1_1    : u32 = VK_MAKE_VERSION(1, 1, 0);

#[cfg(unix)]
const LIB: &str = "libvulkan.so";
#[cfg(windows)]
const LIB: &str = "vulkan-1.dll";

impl std::ops::Try for VkResult {
	type Ok = VkResult;
	type Error = VkResult;
	
	fn into_result(self) -> Result<Self::Ok, Self::Error> {
		match self {
			VK_SUCCESS | VK_NOT_READY | VK_TIMEOUT | VK_EVENT_SET | VK_EVENT_RESET | VK_INCOMPLETE => Ok(self),
			_ => Err(self)
		}
	}
	
	fn from_error(v: Self::Error) -> Self {
		v
	}
	
	fn from_ok(v: Self::Ok) -> Self {
		v
	}
}

impl std::error::Error for VkResult {}

impl std::fmt::Display for VkResult {
	fn fmt(&self, f: &mut Formatter) -> Result<(), Error> {
		std::fmt::Debug::fmt(self, f)
	}
}
"#;

// we have to use macros here, because consts are not constant enough
macro_rules! VK_INIT {
    () => {
    	r#"
static mut LIBVULKAN                                   : Option<libloading::Library> = None;
static mut SYMBOL                                      : Option<libloading::Symbol<extern "system" fn(u64, *const u8) -> PFN_vkVoidFunction>> = None;
static mut pfn_vkGetInstanceProcAddr                   : extern "system" fn(u64, *const u8) -> PFN_vkVoidFunction = {{ extern "system" fn f(arg0: u64, arg1: *const u8) -> PFN_vkVoidFunction {{ unsafe {{ vkInit(); (pfn_vkGetInstanceProcAddr)(arg0, arg1) }} }} f }};{}

unsafe fn vkInit() {{
	use ::std::ops::Deref;
	LIBVULKAN                                   = Some(libloading::Library::new(LIB).expect("failed to load `libvulkan`"));
	SYMBOL                                      = Some(LIBVULKAN.as_ref().unwrap().get(b"vkGetInstanceProcAddr").expect("failed to load `vkGetInstanceProcAddr`"));
	pfn_vkGetInstanceProcAddr                   = *SYMBOL.as_ref().unwrap().deref();{}
}}
{}"#
    };
}

macro_rules! VK_INSTANCE_IMPL {
    () => {
    	r#"
#[derive(Clone)]
pub struct VkInstanceImpl {{
	pub handle: VkInstance{}
}}

impl ::std::ops::Deref for VkInstanceImpl {{
	type Target = VkInstance;
	#[inline]
	fn deref(&self) -> &VkInstance {{ &self.handle }}
}}

impl VkInstanceImpl {{
	pub unsafe fn new(handle: VkInstance) -> Self {{
		VkInstanceImpl {{
			handle{}
		}}
	}}{}}}"#
    };
}

macro_rules! VK_GENERIC_IMPL {
    () => {
    	r#"
/// Wrapper for `{0}`
pub struct {0}Impl<'a> {{
	pub handle: {0},
	pub parent: &'a {1}Impl{2}
}}

impl ::std::ops::Deref for {0}Impl<'_> {{
	type Target = {0};
	
	#[inline]
	fn deref(&self) -> &{0} {{ &self.handle }}
}}

impl ::std::fmt::Debug for {0}Impl<'_> {{
	fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {{
		f.debug_struct("{0}Impl")
			.field("handle", &self.handle)
			.finish()
	}}
}}

impl ::std::fmt::Display for {0}Impl<'_> {{
	fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {{
		write!(f, "{{:#x}}", self.handle)
	}}
}}

impl<'a> {0}Impl<'a> {{
	pub {5}fn new(handle: {0}, parent: &'a {1}Impl{6}) -> Self {{
		Self {{
			handle,
			parent{3}
		}}
	}}{4}}}
"#
    };
}

static NON_DEBUG_STRUCTS: [&str; 11] = [
	"VkPhysicalDeviceProperties",
	"VkExtensionProperties",
	"VkLayerProperties",
	"VkAllocationCallbacks",
	"VkPhysicalDeviceProperties2",
	"VkDebugReportCallbackCreateInfoEXT",
	"VkDebugUtilsMessengerCreateInfoEXT",
	"VkPhysicalDeviceDriverPropertiesKHR",
	"VkPipelineExecutablePropertiesKHR",
	"VkPipelineExecutableStatisticKHR",
	"VkPipelineExecutableInternalRepresentationKHR"
];

fn main() {
	let args: Vec<String> = std::env::args().collect();
	
	if args.len() < 2 {
		panic!("no input file specified")
	}
	
	let file_in = &args[1];
	let file_out = file_in.replace(".xml", ".rs");
	let file_out_cargo = file_in.replace(".xml", ".toml");
	let verbose = args.iter().any(|s| s == "--verbose")
		|| args.iter().any(|s| s == "-v");
	
	println!("parsing file '{}'", file_in);
	println!("output generated code to {}", &file_out);
	println!("output generated cargo.toml to {}", &file_out_cargo);
	
	let file_in = File::open(file_in).expect("failed to open input file");
	
	let mut state = Registry;
	let mut stack = Vec::<String>::new();
	let mut handles = Vec::<String>::new();
	let mut elements = Vec::<VkRegistryElement>::new();
	let mut tmp_enum: Option<String> = None;
	
	let result = parse_xml(file_in, |args| {
		let mut add: Option<String> = None;
		if let Close(tag, _) = &args {
			if !tag.is_empty() && stack.last().unwrap() != *tag {
				panic!("tag was not closed consistently ({} != {})", stack.last().unwrap(), *tag);
			}
			stack.pop();
		} else if let Open(tag, _) = &args {
			add = Some(tag.to_string());
		}
		
		match &mut state {
			Registry => match (stack.last().map(String::as_str), args) {
				(Some("types"), Open("type", attributes)) => match attributes.remove("alias") {
					Some(alias) => elements.push(TypeAlias {
						name: convert_name(attributes.remove("name").unwrap().as_str()),
						feature: None,
						alias,
					}),
					None => match attributes.remove("category").as_ref().map(|s| s.as_str()) {
						cat @ Some("struct") | cat @ Some("union") => state = TypeStruct {
							name:    convert_name(&attributes.remove("name").unwrap()),
							feature: None,
							comment: None,
							r#type:  String::from(cat.unwrap()),
							members: Vec::new()
						},
						Some("basetype") | Some("bitmask") | Some("handle") => state = TypeAlias {
							name:    String::new(),
							feature: None,
							alias:   String::new(),
						},
						Some(category) if verbose => println!("\x1b[33mskipped type: {}\x1b[0m", category),
						None if verbose => println!("\x1b[33mskipped type without category\x1b[0m"),
						_ => ()
					}
				},
				(_, Open("enums", attributes)) => match attributes.remove("alias") {
					Some(alias) => elements.push(TypeAlias {
						name:    convert_name(attributes.remove("name").unwrap().as_str()),
						feature: None,
						alias,
					}),
					None => state = Enums {
						name:    convert_name(attributes.get("name").unwrap()),
						feature: None,
						comment: attributes.remove("comment"),
						r#type:  attributes.remove("type"),
						enums:   Vec::new(),
						aliases: Vec::new()
					}
				}
				(Some("commands"), Open("command", attributes)) => match attributes.remove("alias") {
					Some(alias) => elements.push(CommandAlias {
						name:    convert_name(attributes.remove("name").unwrap().as_str()),
						feature: None,
						alias,
					}),
					None => state = Command {
						name:    String::new(),
						feature: None,
						result:  String::new(),
						comment: attributes.remove("comment"),
						params:  Vec::new()
					}
				}
				(_, Open("comment", _)) => state = Comment(String::new()),
				(_, Open("feature", attributes)) => state = Extension {
					name:      attributes.remove("name").unwrap(),
					number:    0,
					requires:  Vec::new(),
					types:     Vec::new(),
					commands:  Vec::new(),
					enums_new: Vec::new(),
					enums_ext: Vec::new(),
					enums_ext_aliases: Vec::new()
				},
				(_, Open("extension", attributes)) => match attributes.get("supported").map(String::as_str) {
					Some("vulkan") => state = Extension {
						name:      attributes.remove("name").unwrap(),
						number:    attributes.remove("number").map(|v| v.parse::<u32>().unwrap()).unwrap(),
						requires:  attributes.remove("requires").map_or(Vec::new(), |s| s
							.split(',')
							.map(String::from)
							.collect()),
						types:     Vec::new(),
						commands:  Vec::new(),
						enums_new: Vec::new(),
						enums_ext: Vec::new(),
						enums_ext_aliases: Vec::new()
					},
					Some(_) | None => println!("\x1b[33mskipped unsupported extension: '{}'\x1b[0m", attributes.remove("name").unwrap())
				}
				(_, Open("registry", _)) => (),
				(_, Open("types",    _)) => (),
				(_, Open("commands", _)) => (),
				(_, Open(tag,        _)) if verbose => println!("\x1b[33mrejected element: '{}'\x1b[0m", tag),
				_ => ()
			},
			TypeStruct { members, .. } => match (stack.last().unwrap().as_str(), args) {
				(_,        Open("name",     _))            => (),
				(_,        Open("comment",  _))            => (),
				(_,        Open("enum",     _))            => (),
				(_,        Open("type",     _))            => (),
				(_,        Open("member",   _))            => members.push(VkStructMember {
					name:    String::new(),
					r#type:  String::new(),
					comment: None
				}),
				(_,        Open(tag,   _))            => println!("\x1b[33mskipped element in 'type (struct/union)': '{}'\x1b[0m", tag),
				("type",   Close("member",  content)) => members.last_mut().unwrap().r#type = convert_type(
					members.last().unwrap().r#type.as_str(),
					&if let Some(enum_) = &tmp_enum {
						let mut s = content.trim().to_string();
						s.insert_str(1, &enum_);
						tmp_enum = None;
						s
					} else {
						content.to_string()
					}),
				("member", Close("name",    content)) => members.last_mut().unwrap().name.push_str(&convert_name(content)),
				("member", Close("type",    content)) => members.last_mut().unwrap().r#type.push_str(content),
				("member", Close("comment", content)) => if let Some(member) = members.last_mut() { member.comment = Some(String::from(content)) },
				("member", Close("enum",    content)) => tmp_enum = Some(content.trim().to_string().clone()),
				(_,        Close("type",    _))            => {
					elements.push(state.clone());
					state = Registry;
				},
				_ => ()
			}
			TypeAlias { name, alias, .. } => match (stack.last().unwrap().as_str(), args) {
				(_, Open("name", _)) | (_, Open("type", _)) | (_,Open("comment", _)) => (),
				(_,      Open(tag, _)) if verbose => println!("\x1b[33mskipped element in 'type': '{}'\x1b[0m", tag),
				("type", Close("type", content)) => alias.push_str(content),
				(_,      Close("type", content)) => {
					if alias == "VK_DEFINE_HANDLE" { handles.push(name.clone()); }
					*alias = convert_type(alias.as_str(), content);
					elements.push(state.clone());
					state = Registry;
				}
				("type", Close("name", content)) => name.push_str(content),
				("type", Close("comment", _)) => (),
				_ => (),
			}
			Enums { enums, aliases, .. } => match args {
				Open("name",    _) => (),
				Open("type",    _) => (),
				Open("comment", _) => (),
				Open("enum", attributes) => match attributes.remove("alias") {
					Some(alias) => aliases.push(VkEnum {
						name:    attributes.remove("name").unwrap(),
						value:   alias,
						feature: None,
						comment: attributes.remove("comment")
					}),
					None               => enums.push(VkEnum {
						name:    convert_name(attributes.remove("name").unwrap().as_str()),
						value:   if let Some(value) = attributes.remove("value") {
							convert_value(&value)
						} else if let Some(bitpos) = attributes.remove("bitpos") {
							format!("{:#x}", 1 << bitpos.parse::<u32>().unwrap())
						} else {
							if verbose { println!("\x1b[33mskipped enum\x1b[0m"); }
							stack.push(add.unwrap());
							return;
						},
						feature: None,
						comment: attributes.remove("comment")
					})
				}
				Open(tag, _) if verbose => println!("\x1b[33mskipped element in 'enums': '{}'\x1b[0m", tag),
				Close("enums", _) => {
					elements.push(state.clone());
					state = Registry;
				}
				_ => ()
			}
			Command { name, result, params, .. } => match (stack.last().unwrap().as_str(), args) {
				(_,         Open("type", _)) | (_, Open("name", _)) | ("command", Open("proto", _))  => (),
				("command", Open("param", attributes)) => params.push(VkCommandParam {
					name:    String::new(),
					r#type:  String::new(),
					comment: None,
					len:     attributes.remove("len"),
					option:  attributes.remove("optional").map_or(false, |s| s == "true")
				}),
				(_,         Open(tag, _)) if verbose   => println!("\x1b[33mskipped element in 'command': '{}'\x1b[0m", tag),
				("command", Close("proto", content))   => *result = convert_type(result.as_str(), content),
				("command", Close("param", content))   => params.last_mut().unwrap().r#type = convert_type(params.last().unwrap().r#type.as_str(), content),
				("proto",   Close("type",  content))   => result.push_str(content),
				("proto",   Close("name",  content))   => name.push_str(content),
				("param",   Close("type",  content))   => params.last_mut().unwrap().r#type.push_str(content),
				("param",   Close("name",  content))   => params.last_mut().unwrap().name.push_str(&convert_name(content)),
				(_,         Close("command", _)) => {
					elements.push(state.clone());
					state = Registry;
				}
				_ => ()
			}
			Comment(comment) => match args {
				Open(tag, _) if verbose => println!("\x1b[33mskipped element in 'comment': '{}'\x1b[0m", tag),
				Close(tag, content) if tag == "comment" => {
					*comment = String::from(content);
					elements.push(state.clone());
					state = Registry;
				},
				_ => ()
			},
			Extension { number, types, commands, enums_new,
				enums_ext, enums_ext_aliases, .. } => match (stack.last().unwrap().as_str(), args) {
				(_,         Open("require", _)) => (),
				("require", Open("type",    attributes)) => types.push(attributes.remove("name").unwrap()),
				("require", Open("command", attributes)) => commands.push(attributes.remove("name").unwrap()),
				("require", Open("enum",    attributes)) => match attributes.remove("alias") {
					Some(alias) => enums_ext_aliases.push((attributes.remove("name").unwrap(), alias, attributes.remove("extends").unwrap())),
					None => {
						let name = attributes.remove("name").unwrap();
						let value = if let Some(value) = attributes.remove("value") {
							convert_value(&value)
						} else if let Some(bitpos) = attributes.remove("bitpos") {
							format!("{:#x}", 1 << bitpos.parse::<u32>().unwrap())
						} else if let Some(offset) = attributes.remove("offset") {
							format!("{}", 1_000_000_000
								+ (attributes.remove("extnumber").map_or(*number, |v| v.parse::<u32>().unwrap()) as i32 - 1) * 1_000
								+ offset.parse::<i32>().unwrap())
						} else {
							 if verbose { println!("\x1b[33mskipped extension enum\x1b[0m"); }
							stack.push(add.unwrap());
							return;
						};
						
						match attributes.remove("extends") {
							None                 => enums_new.push((name, value)),
							Some(extends) => enums_ext.push((name, value, extends))
						}
					}
				}
				(_, Open(tag, _)) if verbose => println!("\x1b[33mskipped element in 'extension': '{}'\x1b[0m", tag),
				(_, Close("feature", _)) => {
					elements.push(state.clone());
					state = Registry;
				}
				(_, Close("extension", _)) => {
					elements.push(state.clone());
					state = Registry;
				}
				_ => ()
			},
			CommandAlias { .. } | TypePFN { .. } | Macro { .. } => ()
		}
		
		if let Some(tag) = add {
			stack.push(tag);
		}
	});
	
	// TODO generate function pointers, macros and type aliases automatically
	[
		Macro {
			name: "VK_MAKE_VERSION".to_string(),
			content: "(major: u32, minor: u32, patch: u32) -> u32 { (major << 22) | (minor << 12) | patch }".to_string()
		},
		Macro {
			name: "VK_VERSION_MAJOR".to_string(),
			content: "(version: u32) -> u32 { version >> 22 }".to_string()
		},
		Macro {
			name: "VK_VERSION_MINOR".to_string(),
			content: "(version: u32) -> u32 { version >> 12 & 0x3ff }".to_string()
		},
		Macro {
			name: "VK_VERSION_PATCH".to_string(),
			content: "(version: u32) -> u32 { version & 0xfff }".to_string()
		},
		TypePFN {
			name:    "PFN_vkInternalAllocationNotification".to_string(),
			comment: None,
			feature: None,
			result:  "()".to_string(),
			params:  vec![("pUserData".to_string(),       "*const u8".to_string()),
						  ("size".to_string(),            "usize".to_string()),
						  ("allocationType".to_string(),  "VkInternalAllocationType".to_string()),
						  ("allocationScope".to_string(), "VkSystemAllocationScope".to_string())]
		},
		TypePFN {
			name:    "PFN_vkInternalFreeNotification".to_string(),
			comment: None,
			feature: None,
			result:  "()".to_string(),
			params:  vec![("pUserData".to_string(),       "*const u8".to_string()),
						  ("size".to_string(),            "usize".to_string()),
						  ("allocationType".to_string(),  "VkInternalAllocationType".to_string()),
						  ("allocationScope".to_string(), "VkSystemAllocationScope".to_string())]
		},
		TypePFN {
			name:    "PFN_vkReallocationFunction".to_string(),
			comment: None,
			feature: None,
			result:  "()".to_string(),
			params:  vec![("pUserData".to_string(),       "*const u8".to_string()),
						  ("pOriginal".to_string(),       "*const u8".to_string()),
						  ("size".to_string(),            "usize".to_string()),
						  ("alignment".to_string(),       "usize".to_string()),
						  ("allocationScope".to_string(), "VkSystemAllocationScope".to_string())]
		},
		TypePFN {
			name:    "PFN_vkAllocationFunction".to_string(),
			comment: None,
			feature: None,
			result:  "()".to_string(),
			params:  vec![("pUserData".to_string(),       "*const u8".to_string()),
						  ("size".to_string(),            "usize".to_string()),
						  ("alignment".to_string(),       "usize".to_string()),
						  ("allocationScope".to_string(), "VkSystemAllocationScope".to_string())]
		},
		TypePFN {
			name:    "PFN_vkFreeFunction".to_string(),
			comment: None,
			feature: None,
			result:  "()".to_string(),
			params:  vec![("pUserData".to_string(),       "*const u8".to_string()),
						  ("pMemory".to_string(),         "*const u8".to_string())]
		},
		TypePFN {
			name:    "PFN_vkVoidFunction".to_string(),
			comment: None,
			feature: None,
			result:  "()".to_string(),
			params:  Vec::new()
		},
		TypePFN {
			name:    "PFN_vkDebugReportCallbackEXT".to_string(),
			comment: None,
			feature: None,
			result:  "VkBool32".to_string(),
			params:  vec![("flags".to_string(),        "VkDebugReportFlagsEXT".to_string()),
						  ("objectType".to_string(),   "VkDebugReportObjectTypeEXT".to_string()),
						  ("object".to_string(),       "u64".to_string()),
						  ("location".to_string(),     "usize".to_string()),
						  ("messageCode".to_string(),  "i32".to_string()),
						  ("pLayerPrefix".to_string(), "*const u8".to_string()),
						  ("pMessage".to_string(),     "*const u8".to_string()),
						  ("pUserData".to_string(),    "*const u8".to_string())]
		},
		TypePFN {
			name:    "PFN_vkDebugUtilsMessengerCallbackEXT".to_string(),
			feature: Some("VK_EXT_debug_utils".to_string()),
			comment: None,
			result:  "VkBool32".to_string(),
			params:  vec![("messageSeverity".to_string(), "VkDebugUtilsMessageSeverityFlagBitsEXT".to_string()),
						  ("messageTypes".to_string(),    "VkDebugUtilsMessageTypeFlagsEXT".to_string()),
						  ("pCallbackData".to_string(),   "*const VkDebugUtilsMessengerCallbackDataEXT".to_string()),
						  ("pUserData".to_string(),       "*const u8".to_string())]
		}
	].iter().enumerate().for_each(|(i, e)|
		elements.insert(i, e.clone()));
	
	elements.iter_mut()
		.find_map(|e| match e {
			Enums { name, enums, .. } if name == "API Constants" => Some(enums),
			_ => None
		})
		.expect("expecting enum `API Constants`")
		.iter_mut()
		.filter(|e| e.name == "VK_TRUE" || e.name == "VK_FALSE")
		.for_each(|e| e.value.push_str("u32"));
		
	
	match result {
		Ok(_) => println!("\x1b[32mParsing completed.\x1b[0m"),
		Err(_) => println!("\x1b[31mParsing failed.\x1b[0m")
	}
	
	let mut writer = BufWriter::new(OpenOptions::new()
		.write(true)
		.create(true)
		.open(&file_out)
		.expect("failed to open output file"));
	
	match gen(&mut writer, &elements, handles, true) {
		Ok(_) => println!("\x1b[32mCode generation completed.\x1b[0m"),
		Err(_) => println!("\x1b[31mCode generation failed.\x1b[0m")
	}
	
	let mut writer_cargo = BufWriter::new(OpenOptions::new()
		.write(true)
		.create(true)
		.open(&file_out_cargo)
		.expect("failed to open output cargo file"));
	
	match gen_cargo(&mut writer_cargo, &elements) {
		Ok(_) => println!("\x1b[32mCargo.toml generation completed.\x1b[0m"),
		Err(_) => println!("\x1b[31mCargo.toml generation failed.\x1b[0m")
	}
}

#[allow(clippy::single_match)]
/// parses the Vulkan registry from the given [file] and passes the elements to [f]
fn parse_xml<F : FnMut(ParseArgs)>(file: File, mut f : F) -> Result<()> {
	let mut state = ParseContent;
	let mut content = Vec::<String>::new();
	let mut tag = String::new();
	let mut tmp = String::new();
	let mut attribute_name: Option<String> = None;
	let mut attributes = HashMap::<String, String>::new();
	
	for b in BufReader::new(file).bytes() {
		let ch = b? as char;
		match state {
			ParseDocType => match ch {
				'>' => state = ParseContent,
				_   => ()
			}
			ParseTag => match ch {
				' ' | '\t' | '\n' => {
					content.push(String::new());
					state = FindAttributeName;
				}
				'>' => {
					content.push(String::new());
					f(Open(&tag, &mut attributes));
					tag.clear();
					attributes.clear();
					state = ParseContent;
				}
				'/' if tag.is_empty()         => state = ParseClosingTag,
				'?'                           => state = ParseDocType,
				'!'                           => state = ParseComment,
				ch if !XML_INVALID_CHARS.contains(&ch) => tag.push(ch),
				ch => return Err(Error::new(ErrorKind::Other, format!("invalid character in tag name: '{}'", ch)))
			}
			ParseComment => if ch == '>' { state = ParseContent; }
			ParseContent => if ch == '<' {
				state = ParseTag;
			} else if let Some(content) = content.last_mut() {
				content.push(ch);
			}
			ParseClosingTag => match ch {
				'>' => {
					f(Close(&tag, content.pop().map(convert_xml).unwrap_or_else(String::new).as_str()));
					tag.clear();
					state = ParseContent;
				}
				ch if !XML_INVALID_CHARS.contains(&ch) => tag.push(ch),
				ch => return Err(Error::new(ErrorKind::Other, format!("invalid character in closing tag name: '{}'", ch)))
			}
			FindAttributeName => match ch {
				'>' => {
					f(Open(&tag, &mut attributes));
					tag.clear();
					attributes.clear();
					state = ParseContent;
				}
				'/' => {
					f(Open(tag.as_str(), &mut attributes));
					tag.clear();
					attributes.clear();
					state = ParseClosingTag;
				}
				' ' | '\t' | '\n' => continue,
				ch if !XML_INVALID_CHARS.contains(&ch) => {
					tmp.clear();
					tmp.push(ch);
					state = ParseAttributeName;
				}
				ch => return Err(Error::new(ErrorKind::Other, format!("invalid character in attribute name: '{}'", ch)))
			}
			ParseAttributeName => match ch {
				' ' | '=' => {
					attribute_name = Some(convert_xml(tmp));
					tmp = String::new();
					state = FindAttributeValue;
				}
				ch if !XML_INVALID_CHARS.contains(&ch) => tmp.push(ch),
				ch => return Err(Error::new(ErrorKind::Other, format!("invalid character in attribute name: '{}'", ch)))
			}
			FindAttributeValue => match ch {
				'"'       => state = ParseAttributeValue,
				' ' | '=' => (),
				_         => return Err(Error::new(ErrorKind::Other, "invalid attribute assignment"))
			}
			ParseAttributeValue => match ch {
				'"' => {
					attributes.insert(attribute_name.take().unwrap(), convert_xml(tmp));
					tmp = String::new();
					state = FindAttributeName;
				}
				_ => tmp.push(ch)
			}
		}
	}
	
	match state {
		ParseContent => Ok(()),
		_ => Err(Error::new(ErrorKind::Other, "parsing could not be completed"))
	}
}

/// generates rust source code from the given [vec] and writes it to [writer]
fn gen<T: Write + Seek>(writer: &mut T, elements: &[VkRegistryElement], handles: Vec<String>, mangle: bool) -> Result<()> {
	use std::fmt::Write;
	writer.write_all(VK_BEGINNING.as_bytes())?;
	
	let mut instance_table = String::new();
	let mut instance_load  = String::new();
	let mut device_table   = String::new();
	let mut device_load    = String::new();
	let mut static_table   = String::new();
	let mut static_load    = String::new();
	let mut static_cmds    = String::new();
	
	let old = elements;
	let mut elements = Vec::new();
	let mut consts = Enums {
		name:    String::from("extension constants"),
		feature: None,
		comment: None,
		r#type:  None,
		enums:   Vec::new(),
		aliases: Vec::new()
	};
	
	// this works as long as the extensions are located at the end of the registry
	for element in old {
		match element {
			CommandAlias { name, feature, alias } => for e in old {
				if let Command { name: name_, result, params, .. } = e {
					if name_ == alias {
						elements.push(Command {
							name:    name.clone(),
							feature: feature.clone(),
							comment: None,
							result:  result.clone(),
							params:  params.clone()
						});
						break;
					}
				}
			}
			Extension { name: extname, types, commands,
				enums_new, enums_ext, enums_ext_aliases, .. } => {
				for r#enum in enums_new {
					if let Enums { enums, .. } = &mut consts {
						enums.push(VkEnum {
							name:    r#enum.0.clone(),
							value:   r#enum.1.clone(),
							feature: Some(extname.clone()),
							comment: None
						});
					}
				}
				
				for e in &mut elements {
					match e {
						TypeAlias    { name, feature, .. }
						| TypeStruct { name, feature, .. }
						| TypePFN    { name, feature, .. } => for r#type in types {
							if name == r#type {
								*feature = Some(extname.clone());
								break;
							}
						}
						Command { name, feature, .. } => for command in commands {
							if name == command {
								*feature = Some(extname.clone());
								break;
							}
						}
						Enums { name, enums, aliases, .. } => {
							for (name_, value, extends) in enums_ext {
								if name == extends && enums.iter().find(|e| &e.name == name_).is_none() {
									enums.push(VkEnum {
										name:    name_.clone(),
										value:   value.clone(),
										feature: Some(extname.clone()),
										comment: None
									});
								}
							}
							
							for (name_, alias, extends) in enums_ext_aliases {
								if name == extends {
									aliases.push(VkEnum {
										name:    name_.clone(),
										value:   alias.clone(),
										feature: Some(extname.clone()),
										comment: None
									});
								}
							}
						}
						_ => ()
					}
				}
			}
			_ => elements.push(element.clone())
		}
	}
	
	elements.push(consts);
	
	let mut buffers = HashMap::<String, String>::new();
	for handle in handles {
		buffers.insert(handle, String::new());
	}
	
	for element in &elements {
		match element {
			Comment(comment) => for l in comment.lines() { writeln!(writer, "// {}", l)?; },
			TypeAlias { name, feature, alias } => {
				if let Some(feature) = feature { writeln!(writer, "#[cfg(feature = \"{}\")]", feature)?; }
				writeln!(writer, "pub type {:<45} = {};", name, convert_type(&alias, ""))?;
			},
			TypeStruct { name, feature, comment, r#type, members } => {
				writeln!(writer)?;
				if let Some(comment) = comment {
					for l in comment.lines() {
						writeln!(writer, "/// {}", l)?;
					}
				}
				
				if let Some(feature) = feature { writeln!(writer, "#[cfg(feature = \"{}\")]", feature)?; }
				
				match r#type.as_str() {
					"struct" => writeln!(writer, "#[repr(C)]\n#[derive(Copy, Clone{})]\npub struct {} {{",
							if NON_DEBUG_STRUCTS.contains(&name.as_str()) { "" } else { ", Debug" }, name),
					"union"  => writeln!(writer, "#[repr(C)]\n#[derive(Copy, Clone)]\npub union {} {{", name),
					_ => unreachable!()
				}?;
				
				for VkStructMember { name, r#type, comment } in members {
					if let Some(comment) = comment { writeln!(writer, "\t/// {}", comment)?; }
					writeln!(writer, "\tpub {:<45}: {},", name, r#type)?;
				}
				
				writeln!(writer, "}}\n")?;
				
				if r#type == "union" {
					if let Some(feature) = feature { writeln!(writer, "#[cfg(feature = \"{}\")]", feature)?; }
					writeln!(writer, "impl Debug for {} {{ fn fmt(&self, f: &mut Formatter) -> Result<(), Error> {{ f.write_str(\"{0} {{ }}\") }} }}\n", name)?;
				}
				
				if let Some(feature) = feature { writeln!(writer, "#[cfg(feature = \"{}\")]", feature)?; }
				writeln!(writer, "impl Default for {} {{ fn default() -> Self {{ unsafe {{ MaybeUninit::zeroed().assume_init() }} }} }}", name)?;
			}
			TypePFN { name, feature, comment, result, params } => {
				if let Some(comment) = comment {
					for l in comment.lines() {
						write!(writer, "\n/// {}", l)?;
					}
				}
				
				if let Some(feature) = feature {
					write!(writer, "\n#[cfg(feature = \"{}\")]", feature)?;
				}
				
				write!(writer, "\npub type {} = extern \"system\" fn(", name)?;
				
				for (name, r#type) in params {
					write!(writer, "\n\t{:<45}: {},", name, r#type)?;
				}
				
				if !params.is_empty() { writer.seek(SeekFrom::Current(-1)).unwrap(); }
				write!(writer, ")\n\t-> {};\n", result)?;
			}
			Enums { name, comment, r#type, enums, aliases, .. } => {
				if let Some(comment) = comment {
					for l in comment.lines() {
						write!(writer, "\n/// {}", l)?;
					}
				}
				
				let name_ = name;
				match *r#type {
					Some(_) => {
						let mut has_features = false;
						for e in enums.iter() {
							match e.feature.as_ref() {
								Some(_) => has_features = true,
								None => { has_features = false; break; }
							}
						}
						
						let features = if has_features {
							format!("#[cfg(any({}))]\n", enums.iter()
								.map(|e| e.feature.as_ref())
								.filter_map(identity)
								.collect::<HashSet<&String>>()
								.iter()
								.map(|feature| format!("{}, ", feature))
								.collect::<String>()
								.trim_end_matches(", "))
						} else {
							String::new()
						};
						
						writeln!(writer, "\n{1}pub use self::{0}::*;\n{2}\n#[derive(Copy, Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]\n{1}pub enum {0} {{",
								 name, features, if enums.is_empty() { "" } else { "\n#[repr(C)]" })?;
						
						for VkEnum { name, value, feature, comment } in enums {
							if let Some(comment) = comment { writeln!(writer, "\t/// {}", comment)?; }
							if let Some(feature) = feature { writeln!(writer, "\t#[cfg(feature = \"{}\")]", feature)?; }
							
							/*let mut name = name.clone();
							if mangle {
								let mut buf = String::new();
								let mut iter = name.chars();
								
								loop {
									match iter.next() {
										None => break,
										Some('_') => if let Some(ch) = iter.next() { buf.push(ch.to_ascii_uppercase()) }
										Some(ch) if buf.is_empty() => buf.push(ch.to_ascii_uppercase()),
										Some(ch) => buf.push(ch.to_ascii_lowercase())
									}
								}
								
								name = String::from(buf.trim_start_matches(name_));
								
								if name.chars().next().unwrap().is_ascii_digit() {
									name.insert(0, '_');
								}
							}*/
							
							writeln!(writer, "\t{:<80} = {},", name, value)?;
						}
						
						writeln!(writer, "}}")?;
						
						if !aliases.is_empty() { writeln!(writer)?; }
						for VkEnum { name, value: alias, feature, comment } in aliases {
							if let Some(comment) = comment { writeln!(writer, "/// {}", comment)?; }
							if let Some(feature) = feature { writeln!(writer, "#[cfg(feature = \"{}\")]", feature)?; }
							writeln!(writer, "pub const {:<80}: {} = {};", name, name_, alias)?;
						}
					}
					None => {
						writeln!(writer)?;
						
						for VkEnum { name, value, feature, comment } in enums {
							if let Some(comment) = comment { writeln!(writer, "/// {}", comment)?; }
							if let Some(feature) = feature { writeln!(writer, "#[cfg(feature = \"{}\")]", feature)?; }
							writeln!(writer, "pub const {:<50}: {} = {};", name, parse_type(value), value)?;
						}
					}
				}
			}
			Macro { name, content } => writeln!(writer, "\n#[inline(always)] pub const fn {}{}", name, content)?,
			Command { name, feature, comment, result, params } => {
				
				// skip loading functions
				match name.as_str() {
					"vkGetInstanceProcAddr" | "vkGetDeviceProcAddr" => continue,
					_ => ()
				}
				
				// get dispatchable handle
				let handle = params.get(0)
					.map(|param| &r#param.r#type)
					.filter(|handle| buffers.contains_key(*handle))
					.map(String::as_str);
				
				// mangle
				let mut name2 = name.clone();
				if mangle {
					if let Some(handle) = handle {
						name2 = name.replace(&handle[2..], "")
					}
					// remove `vk`
					name2.remove(0);
					name2.remove(0);
					// fist char to lower case
					let ch = name2.remove(0).to_ascii_lowercase();
					name2.insert(0, ch);
				}
				
				// cmds:    buffer that contains the actual methods
				// table:   the table that contains the function pointers
				// init:    contains calls to load the functions into the table
				// load_fn: the method used to load this function
				let (cmds, table, init, load_fn) = match handle {
					Some(handle @ "VkInstance") | Some(handle @ "VkPhysicalDevice")
					                    => (buffers.get_mut(handle).unwrap(), &mut instance_table, &mut instance_load, "pfn_vkGetInstanceProcAddr"),
					Some(handle)	=> (buffers.get_mut(handle).unwrap(), &mut device_table,   &mut device_load,   "parent.pfn_vkGetDeviceProcAddr"),
					None         		=> (&mut static_cmds,                 &mut static_table,   &mut static_load,   "pfn_vkGetInstanceProcAddr"),
				};
				
				// get dispatcher in whose table this function should be contained
				let dispatcher = match handle {
					None => "unsafe { &r#static }",
					Some("VkInstance") | Some("VkDevice") => "self",
					Some(_) => "self.parent"
				};
				
				// create some buffers
				let mut _tmp; // tmp var due to lifetime rules
				let mut proto = String::new();
				let mut header = String::new();
				let mut call = String::new();
				let mut lens = HashMap::new();
				
				// init buffers
				if let Some(handle) = &handle {
					proto.push_str(handle);
					proto.push_str(", ");
					header.push_str("\n\t&self,");
					call.push_str("self.handle, ");
				}
				
				// get array lengths
				for param in &params[if handle.is_none() { 0 } else { 1 }..] {
					match &param.len {
						// filter optional params, so that it's possible to pass a null ptr
						Some(len) if param.r#type.starts_with("*const ") && !param.option && !lens.contains_key(len) => {
							lens.insert(len, param);
						}
						_ => ()
					}
				}
				
				enum ParamType { DEF, ARR, LEN }
				
				// write buffers
				for VkCommandParam { name, r#type, comment, len, .. } in &params[if handle.is_none() { 0 } else { 1 }..] {
					let param_type = match len.as_ref() {
						Some(len) if lens.contains_key(len) && r#type.starts_with("*const ") => ParamType::ARR,
						_ if lens.contains_key(&name) => ParamType::LEN,
						_ => ParamType::DEF,
					};
					let name = lens.get(&name)
						.map_or(name, |param| &param.name);
					
					proto.push_str(&r#type);
					proto.push_str(", ");
					call.push_str(&name);
					call.push_str(match param_type {
						ParamType::LEN => { _tmp = format!(".len() as {}, ", r#type); &_tmp }
						ParamType::ARR => ".as_ptr(), ",
						ParamType::DEF => ", "
					});
					
					if let Some(comment) = comment { write!(header, " // {}", comment).unwrap(); }
					
					match param_type {
						ParamType::ARR => write!(header, "\n\t{:<45}: &[{}],", name, &r#type["*const ".len()..]),
						ParamType::DEF => write!(header, "\n\t{:<45}: {},", name, r#type),
						ParamType::LEN => continue
					}.unwrap();
				};
				
				// write comments
				if let Some(comment) = comment {
					for l in comment.lines() {
						cmds.push_str(&format!("\n/// {}", l));
					}
				}
				
				// remove tailing colon
				let proto = proto.trim_end_matches(", ");
				let header = header.trim_end_matches(',');
				let call = call.trim_end_matches(", ");
				
				if handle.is_some() {
					table.push(',');
					init.push(',');
					
					// write feature
					if let Some(feature) = feature {
						write!(cmds, "\n\t#[cfg(feature = \"{}\")]", feature).unwrap();
						write!(table, "\n\t#[cfg(feature = \"{}\")]", feature).unwrap();
						write!(init, "\n\t\t\t#[cfg(feature = \"{}\")]", feature).unwrap();
					}
					
					// write command, table entry and loading code
					write!(cmds, "{}", format!("\n#[inline]\npub fn {}({}\n) -> {} {{\n\t({}.pfn_{})({})\n}}\n", name2, &header, result, dispatcher, name, &call).replace("\n", "\n\t")).unwrap();
					write!(table, "\n\tpfn_{:<60}: extern \"system\" fn({}) -> {}", name, proto, result).unwrap();
					write!(init, "\n\t\t\tpfn_{:<60}: transmute(({})(handle, b\"{0}\\0\".as_ptr()))", name, load_fn).unwrap();
				} else {
					write!(cmds, "\n#[inline]\npub fn {}({}\n) -> {} {{\n\tunsafe {{ (pfn_{})({}) }}\n}}\n", name2, &header, result, name, &call).unwrap();
					write!(table, "\nstatic mut pfn_{0:<40}: extern \"system\" fn({1}) -> {2} = {{extern \"system\" fn f({3}) -> {2} {{ unsafe {{ vkInit(); (pfn_{0})({4}) }} }} f }};",
						   name, proto, result, header.lines().map(str::trim).collect::<String>(), &call).unwrap();
					write!(init, "\n\tpfn_{:<40}= transmute((pfn_vkGetInstanceProcAddr)(VK_NULL_HANDLE, b\"{0}\\0\".as_ptr()));", name).unwrap();
				}
			}
			_ => ()
		}
	}
	
	instance_table.push_str(",pfn_vkGetDeviceProcAddr                                    : extern \"system\" fn(VkDevice, *const u8) -> PFN_vkVoidFunction");
	instance_load.push_str(",pfn_vkGetDeviceProcAddr                                         : transmute((pfn_vkGetInstanceProcAddr)(handle, b\"vkGetDeviceProcAddr\\0\".as_ptr()))");
	writeln!(writer, VK_INIT!(), static_table, static_load, static_cmds)?;
	
	for (handle, commands) in buffers {
		match handle.as_str() {
			"VkPhysicalDevice" => write!(writer, VK_GENERIC_IMPL!(), "VkPhysicalDevice", "VkInstance", "", "", &commands, "", ""),
			"VkDevice"         => write!(writer, VK_GENERIC_IMPL!(), "VkDevice", "VkInstance", &device_table, &device_load, &commands, if device_table.is_empty() { "" } else { "unsafe " }, ""),
			"VkQueue"          => write!(writer, VK_GENERIC_IMPL!(), "VkQueue", "VkDevice", "<'a>", "", &commands, "", "<'a>"),
			"VkCommandBuffer"  => write!(writer, VK_GENERIC_IMPL!(), "VkCommandBuffer", "VkDevice", "<'a>", "", &commands, "", "<'a>"),
			"VkInstance"       => writeln!(writer, VK_INSTANCE_IMPL!(), instance_table, instance_load, commands),
			handle => panic!("undefined handle type: {}", handle)
		}?;
	}
	Ok(())
}

fn gen_cargo<T: Write>(writer: &mut T, elements: &[VkRegistryElement]) -> Result<()> {
	writeln!(writer, "[dependencies]\nlibloading = \"0.*\"\n")?;
	writeln!(writer, "[features]\ndefault = [\"VK_VERSION_1_0\", \"VK_VERSION_1_1\"]")?;
	let mut extensions = elements.iter()
		.filter_map(|e| match e {
			Extension { name, requires, .. } => Some((name, requires)),
			_ => None
		})
		.collect::<Vec<_>>();
	extensions.sort_by_key(|v| v.0);
	
	for (name, requires) in extensions {
		use std::fmt::Write;
		let mut dependencies = String::new();
		for s in requires { write!(&mut dependencies, "\"{}\", ", s).unwrap(); }
		writeln!(writer, "{:<50}= [{}]", name, dependencies.trim_end_matches(", "))?;
	}
	
	Ok(())
}

/// contains either the data of an opening tag or a closing tag
#[derive(Debug)]
enum ParseArgs<'a> {
	Open(&'a str, &'a mut HashMap<String, String>),
	Close(&'a str, &'a str)
}

/// represents the state of the XML parser
#[derive(PartialEq, Eq, Debug)]
enum State {
	ParseDocType,
	ParseComment,
	ParseContent,
	ParseTag,
	ParseClosingTag,
	FindAttributeName,
	ParseAttributeName,
	FindAttributeValue,
	ParseAttributeValue
}

/// represents an element of the Vulkan registry
#[derive(PartialEq, Eq, Debug, Clone)]
enum VkRegistryElement {
	Registry,
	TypeAlias {
		name:    String,
		feature: Option<String>,
		alias:   String,
	},
	TypeStruct {
		name:    String,
		feature: Option<String>,
		comment: Option<String>,
		r#type:  String,
		members: Vec<VkStructMember>,
	},
	TypePFN {
		name:    String,
		feature: Option<String>,
		comment: Option<String>,
		result:  String,
		params:  Vec<(String, String)>,
	},
	Enums {
		name:    String,
		feature: Option<String>,
		comment: Option<String>,
		r#type:  Option<String>,
		enums:   Vec<VkEnum>,
		aliases: Vec<VkEnum>,
	},
	Command {
		name:    String,
		feature: Option<String>,
		comment: Option<String>,
		result:  String,
		params:  Vec<VkCommandParam>,
	},
	CommandAlias {
		name:    String,
		feature: Option<String>,
		alias:   String,
	},
	Macro {
		name:    String,
		content: String,
	},
	Comment(String),
	Extension {
		name:              String,
		number:            u32,
		requires:          Vec<String>,
		types:             Vec<String>,
		commands:          Vec<String>,
		enums_new:         Vec<(String, String)>,
		enums_ext:         Vec<(String, String, String)>,
		enums_ext_aliases: Vec<(String, String, String)>,
	},
}

#[derive(PartialEq, Eq, Debug, Clone)]
struct VkStructMember {
	name:    String,
	r#type:  String,
	comment: Option<String>
}

#[derive(PartialEq, Eq, Debug, Clone)]
struct VkEnum {
	name:    String,
	value:   String,
	feature: Option<String>,
	comment: Option<String>
}

#[derive(PartialEq, Eq, Debug, Clone)]
struct VkCommandParam {
	name:    String,
	r#type:  String,
	comment: Option<String>,
	len:     Option<String>,
	option:  bool
}

static XML_EXP: [(&str, char); 9] = [
	("&exclamation;", '!'),
	("&quot;",        '"'),
	("&percent;",     '%'),
	("&amp;",         '&'),
	("&apos;",        '\''),
	("&add;",         '+'),
	("&lt;",          '<'),
	("&equal;",       '='),
	("&gt;",          '>')
];

static XML_INVALID_CHARS: [char; 29] = [
	'!', '"', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',',  '/', ';', '<',
	'=', '>', '?', '@', '[', '\\', ']', '^', '`', '{', '|', '}', '~', ' '
];

static C_TYPES: [(&str, &str); 16] = [
	("int8_t",   "i8"),
	("int16_t",  "i16"),
	("int32_t",  "i32"),
	("int64_t",  "i64"),
	("uint8_t",  "u8"),
	("uint16_t", "u16"),
	("uint32_t", "u32"),
	("uint64_t", "u64"),
	("float",    "f32"),
	("double",   "f64"),
	("size_t",   "usize"),
	("int",      "i32"),
	("char",     "u8"),
	("void",     "()"),
	("VK_DEFINE_HANDLE", "u64"),
	("VK_DEFINE_NON_DISPATCHABLE_HANDLE", "u64")
];

static KEYWORDS: [&str; 50] = [
	"as",
	"break",
	"const",
	"continue",
	"crate",
	"dyn",
	"else",
	"enum",
	"extern",
	"false",
	"fn",
	"for",
	"if",
	"impl",
	"in",
	"let",
	"loop",
	"match",
	"mod",
	"move",
	"mut",
	"pub",
	"ref",
	"return",
	"Self",
	"self",
	"static",
	"struct",
	"super",
	"trait",
	"true",
	"type",
	"unsafe",
	"use",
	"where",
	"while",
	"abstract",
	"async",
	"become",
	"box",
	"do",
	"final",
	"macro",
	"override",
	"priv",
	"try",
	"typeof",
	"unsized",
	"virtual",
	"yield"
];

/// converts xml expressions (e.g.: &quot; -> ")
fn convert_xml(mut s: String) -> String {
	for (k, v) in XML_EXP.iter() {
		s = s.replace(k, v.encode_utf8(&mut [0u8; 1]));
	}
	s
}

/// converts C-types to Rust-types (e.g. uint32_t -> u32)
fn convert_type(type0: &str, type1: &str) -> String {
	#[allow(clippy::never_loop)]
	let type0 = 'outer: loop {
		for (k, v) in C_TYPES.as_ref() {
			if *k == type0 { break 'outer *v; }
		}
		break type0;
	};
	
	let type1 = match type1.trim() {
		"*"        | "struct *"        => "*mut",
		"**"       | "struct **"       => "*mut *mut",
		"const *"  | "const struct *"  => "*const",
		"const **" | "const struct **" => "*mut *const",
		"const * const*"               => "*const *const",
		"" | "typedef  ;" | "()"       => return type0.to_string(),
		r#type                    => {
			let r#type = r#type.trim_start_matches("const  ");
			if r#type.starts_with('[') && r#type.ends_with(']') {
				return format!("[{}; {} as usize]", type0, &r#type[1..r#type.len() - 1]);
			}
			panic!("undefined pointer type: {}", r#type);
		},
	};
	let mut s = String::new();
	s.push_str(type1);
	s.push_str(" ");
	s.push_str(if !type1.is_empty() && type0 == "()" { "u8" } else { type0 });
	s
}

/// converts C-style values to Rust-style values (e.g.: 1ULL -> 1u64)
fn convert_value(name: &str) -> String {
	if name.starts_with('\"') && name.ends_with('\"') {
		let mut name = String::from(name);
		name.insert(0, 'b');
		name.truncate(name.len() - 1);
		name.push_str("\\0\".as_ptr()");
		return name;
	}
	
	let mut s = String::new();
	let mut i = 0;
	let name = name.as_bytes();
	
	while i < name.len() {
		if name[i] >= b'0' && name[i] <= b'9' {
			s.push(name[i] as char);
			if name[i + 1..].starts_with(b"ULL") {
				s.push_str("u64");
				i += 3;
			} else if i + 1 < name.len() && name[i + 1] == b'U' {
				s.push_str("u32");
				i += 1;
			} else if i + 1 < name.len() && name[i + 1] == b'f' {
				s.push_str("f32");
				i += 1;
			}
		} else if i + 1 < name.len() && name[i] == b'~' && name[i + 1] >= b'0' && name[i + 1] <= b'9'{
			s.push('!');
		} else {
			s.push(name[i] as char);
		}
		i += 1;
	}
	s
}

/// converts names that contain Rust keywords (e.g.: type -> r#type)
fn convert_name(name: &str) -> String {
	match name {
		name if KEYWORDS.contains(&name) => format!("r#{}", name),
		name => name.to_string()
	}
}

/// parses the type of the given value (e.g. 1u32 -> u32) or returns a default type (i32)
fn parse_type(s: &str) -> &str {
	if      s.contains("u128")   { "u128" }
	else if s.contains("u64")    { "u64"  }
	else if s.contains("u32")    { "u32"  }
	else if s.contains("u16")    { "u16"  }
	else if s.contains("u8")     { "u8"   }
	else if s.contains("i128")   { "i128" }
	else if s.contains("i64")    { "i64"  }
	else if s.contains("i32")    { "i32"  }
	else if s.contains("i16")    { "i16"  }
	else if s.contains("i8")     { "i8"   }
	else if s.contains("f64")    { "f64"  }
	else if s.contains("f32")    { "f32"  }
	else if s.contains("usize")  { "usize" }
	else if s.contains("isize")  { "isize" }
	else if s.contains("char")   { "char" }
	else if s.starts_with('\"') && s.ends_with('\"') { "&str" }
	else if s.starts_with("b\"") && s.ends_with("\\0\".as_ptr()") { "*const u8" }
	else { "i32" }
}
